import sys
from PyQt4 import QtCore, QtGui
import PyQt4.QtGui as QtWidgets

from AboutDialog import AboutDialog
from Classifier import Classifier
from ClassifierParameters import ClassifierParameters
from Clustering import Clustering
from ClusteringParameters import ClusteringParameters
from CreateSamplesProperties import CreateSamplesProperties
from DimensionalityReduction import DimensionalityReduction
from DimensionalityReductionParameters import DimensionalityReductionParameters
from FeatureSpace import FeatureSpace
from Parameters import resource_path
from Regression import Regression
from RegressionParameters import RegressionParameters


class PyClassificationToolbox(QtWidgets.QMainWindow):

	def __init__(self):
		super(PyClassificationToolbox, self).__init__()

		self.initUI()

		try:
			self.featurespace.loadDefaultFeatureSpace('.featurespace.pyct')
		except:
			print("could not open default feature space")


	def initUI(self):
		self.clusteringParameters = ClusteringParameters(self)
		self.dimRedParameters = DimensionalityReductionParameters(self)
		self.classifierParameters = ClassifierParameters(self)
		self.regressionParameters = RegressionParameters(self)
		self.aboutDialog = AboutDialog(self)

		self.createSamplesDockWidget = CreateSamplesProperties(self)
		self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.createSamplesDockWidget)

		self.statusbar = QtWidgets.QStatusBar()
		self.setStatusBar(self.statusbar)

		self.featurespace = FeatureSpace(self, self.statusbar, self.createSamplesDockWidget)		
		self.setCentralWidget(self.featurespace)
		
		self.classifier = None
		self.regressor = None

		
		featureSpaceNewAction = QtWidgets.QAction('&New', self)
		featureSpaceNewAction.setShortcut('Ctrl+N')
		featureSpaceNewAction.setStatusTip('Create empty feature space')
		featureSpaceNewAction.triggered.connect(self.newFeatureSpace)

		featureSpaceOpenAction = QtWidgets.QAction('&Open...', self)
		featureSpaceOpenAction.setShortcut('Ctrl+O')
		featureSpaceOpenAction.setStatusTip('Load a feature space')
		featureSpaceOpenAction.triggered.connect(self.openFeatureSpace)

		featureSpaceSaveAction = QtWidgets.QAction('&Save', self)
		featureSpaceSaveAction.setShortcut('Ctrl+S')
		featureSpaceSaveAction.setStatusTip('Save the feature space')
		featureSpaceSaveAction.triggered.connect(self.saveFeatureSpace)

		featureSpaceSaveAsAction = QtWidgets.QAction('Save &as...', self)
		featureSpaceSaveAsAction.setStatusTip('Save the feature space to a new file')
		featureSpaceSaveAsAction.triggered.connect(self.saveAsFeatureSpace)

		featureSpaceImportAction = QtWidgets.QAction('&Import samples...', self)
		featureSpaceImportAction.setStatusTip('Read feature vectors from an ASCII file')
		featureSpaceImportAction.triggered.connect(self.importFeatureSpace)

		featureSpaceExportAction = QtWidgets.QAction('&Export samples...', self)
		featureSpaceExportAction.setStatusTip('Write the feature vectors to an ASCII file')
		featureSpaceExportAction.triggered.connect(self.exportFeatureSpace)
		
		featureSpaceSaveImageAction = QtWidgets.QAction('Export as image...', self)
		featureSpaceSaveImageAction.setStatusTip('Export the feature space as image')
		featureSpaceSaveImageAction.triggered.connect(self.exportFeatureSpaceAsImage)

		exitAction = QtWidgets.QAction('&Quit', self)
		exitAction.setShortcut(QtCore.Qt.CTRL + QtCore.Qt.Key_Q)
		exitAction.setStatusTip('Exit the Python Classification Toolbox')
		exitAction.triggered.connect(self.close)

		menubar = self.menuBar()
		menubar.setNativeMenuBar(False)
		featureSpaceMenu = menubar.addMenu('&Feature Space')
		featureSpaceMenu.addAction(featureSpaceNewAction)
		featureSpaceMenu.addAction(featureSpaceOpenAction)
		featureSpaceMenu.addAction(featureSpaceSaveAction)
		featureSpaceMenu.addAction(featureSpaceSaveAsAction)
		featureSpaceMenu.addSeparator()
		featureSpaceMenu.addAction(featureSpaceImportAction)
		featureSpaceMenu.addAction(featureSpaceExportAction)
		featureSpaceMenu.addAction(featureSpaceSaveImageAction)
		featureSpaceMenu.addSeparator()
		featureSpaceMenu.addAction(exitAction)


		clusteringKMeansAction = QtWidgets.QAction('k-Means Clustering', self)
		clusteringKMeansAction.setStatusTip('k-Means Clustering')
		clusteringKMeansAction.triggered.connect(lambda: self.clusterWithParameters(Clustering.kMeans))

		clusteringGMMAction = QtWidgets.QAction('Gaussian Mixture Model', self)
		clusteringGMMAction.setStatusTip('Gaussian Mixture Model')
		clusteringGMMAction.triggered.connect(lambda: self.clusterWithParameters(Clustering.GMM))

		clusteringParametersAction = QtWidgets.QAction('Parameters...', self)
		clusteringParametersAction.setStatusTip('Edit the parameters of the clustering algorithms')
		clusteringParametersAction.triggered.connect(self.editClusteringParameters)
		
		clusteringMenu = menubar.addMenu('C&lustering')
		clusteringMenu.addAction(clusteringKMeansAction)
		clusteringMenu.addAction(clusteringGMMAction)
		clusteringMenu.addSeparator()
		clusteringMenu.addAction(clusteringParametersAction)


		dimRedPCAAction = QtWidgets.QAction('Principal Component Analysis', self)
		dimRedPCAAction.setStatusTip('Principal Component Analysis (PCA)')
		dimRedPCAAction.triggered.connect(lambda: self.reduceDimensionalityWithParameters(DimensionalityReduction.PCA))

		dimRedMenu = menubar.addMenu('&Dimensionality Reduction')
		dimRedMenu.addAction(dimRedPCAAction)


		classificationLogRegAction = QtWidgets.QAction('Linear Logistic Regression...', self)
		classificationLogRegAction.setStatusTip('Linear Logistic Regression classifier')
		classificationLogRegAction.triggered.connect(lambda: self.classifyWithParameters(Classifier.LogReg))

		classificationNormAction = QtWidgets.QAction('Norm classifier...', self)
		classificationNormAction.setStatusTip('Classification based on the distance to the class centers')
		classificationNormAction.triggered.connect(lambda: self.classifyWithParameters(Classifier.Norm))

		classificationNaiveBayesAction = QtWidgets.QAction('Naive Bayes', self)
		classificationNaiveBayesAction.setStatusTip('Naive Bayes classifier')
		classificationNaiveBayesAction.triggered.connect(lambda: self.classify(Classifier.NaiveBayes))

		classificationGaussianAction = QtWidgets.QAction('Gaussian classifier', self)
		classificationGaussianAction.setStatusTip('Gaussian classifier')
		classificationGaussianAction.triggered.connect(lambda: self.classify(Classifier.Gauss))

		classificationGMMAction = QtWidgets.QAction('GMM classifier...', self)
		classificationGMMAction.setStatusTip('Gaussian Mixture Model classifier')
		classificationGMMAction.triggered.connect(lambda: self.classifyWithParameters(Classifier.GMM))

		classificationKNNAction = QtWidgets.QAction('kNN...', self)
		classificationKNNAction.setStatusTip('k Nearest Neighbor classifier')
		classificationKNNAction.triggered.connect(lambda: self.classifyWithParameters(Classifier.kNN))

		classificationLinRegAction = QtWidgets.QAction('Linear Regression...', self)
		classificationLinRegAction.setStatusTip('Linear Regression classifier')
		classificationLinRegAction.triggered.connect(lambda: self.classifyWithParameters(Classifier.LinReg))
		
		classificationPerceptronAction = QtWidgets.QAction('Rosenblatt\'s Perceptron...', self)
		classificationPerceptronAction.setStatusTip('Rosenblatt\'s Perceptron')
		classificationPerceptronAction.triggered.connect(lambda: self.classifyWithParameters(Classifier.Perceptron))

		classificationMLPAction = QtWidgets.QAction('Multilayer Perceptron...', self)
		classificationMLPAction.setStatusTip('Multilayer Perceptron')
		classificationMLPAction.triggered.connect(lambda: self.classifyWithParameters(Classifier.MLP))
		classificationMLPAction.setEnabled(False)

		classificationSVMAction = QtWidgets.QAction('SVM...', self)
		classificationSVMAction.setStatusTip('Support Vector Machine')
		classificationSVMAction.triggered.connect(lambda: self.classifyWithParameters(Classifier.SVM))
		
		classificationDecisionForestAction = QtWidgets.QAction('Decision Forest...', self)
		classificationDecisionForestAction.setStatusTip('Decision forest')
		classificationDecisionForestAction.triggered.connect(lambda: self.classifyWithParameters(Classifier.DecForest))

		classificationParametersAction = QtWidgets.QAction('Parameters...', self)
		classificationParametersAction.setStatusTip('Edit the parameters of the classification algorithms')
		classificationParametersAction.triggered.connect(self.editClassificationParameters)

		classificationNoneAction = QtWidgets.QAction('None', self)
		classificationNoneAction.setStatusTip('Delete classification results')
		classificationNoneAction.triggered.connect(self.unsetClassifier)

		classificationMenu = menubar.addMenu('&Classification')
		classificationMenu.addAction(classificationLogRegAction)
		classificationMenu.addAction(classificationNormAction)
		classificationMenu.addAction(classificationNaiveBayesAction)
		classificationMenu.addAction(classificationGaussianAction)
		classificationMenu.addAction(classificationGMMAction)
		classificationMenu.addAction(classificationKNNAction)
		classificationMenu.addAction(classificationLinRegAction)
		classificationMenu.addAction(classificationPerceptronAction)
		classificationMenu.addAction(classificationMLPAction)
		classificationMenu.addAction(classificationSVMAction)
		classificationMenu.addAction(classificationDecisionForestAction)
		classificationMenu.addSeparator()
		classificationMenu.addAction(classificationParametersAction)
		classificationMenu.addAction(classificationNoneAction)


		regressionLinRegAction = QtWidgets.QAction('Linear Regression', self)
		regressionLinRegAction.setStatusTip('Linear Regression')
		regressionLinRegAction.triggered.connect(lambda: self.regressionWithParameters(Regression.LinearRegression))

		regressionSVRAction = QtWidgets.QAction('Support Vector Regression', self)
		regressionSVRAction.setStatusTip('Support Vector Regression (SVR)')
		regressionSVRAction.triggered.connect(lambda: self.regressionWithParameters(Regression.SVR))
		
		regressionRegForestAction = QtWidgets.QAction('Regression Forest', self)
		regressionRegForestAction.setStatusTip('Regression Forest')
		regressionRegForestAction.triggered.connect(lambda: self.regressionWithParameters(Regression.RegForest))

		regressionParametersAction = QtWidgets.QAction('Parameters...', self)
		regressionParametersAction.setStatusTip('Edit the parameters of the regression algorithms')
		regressionParametersAction.triggered.connect(self.editRegressionParameters)

		regressionNoneAction = QtWidgets.QAction('None', self)
		regressionNoneAction.setStatusTip('Delete regression result')
		regressionNoneAction.triggered.connect(self.unsetRegressor)

		regressionMenu = menubar.addMenu('&Regression')
		regressionMenu.addAction(regressionLinRegAction)
		regressionMenu.addAction(regressionSVRAction)
		regressionMenu.addAction(regressionRegForestAction)
		regressionMenu.addSeparator()
		regressionMenu.addAction(regressionParametersAction)
		regressionMenu.addAction(regressionNoneAction)


		aboutAction = QtWidgets.QAction('About...', self)
		aboutAction.setStatusTip('About this software')
		aboutAction.triggered.connect(self.aboutDialog.exec_)

		helpMenu = menubar.addMenu('&Help')
		helpMenu.addAction(aboutAction)

#		exitAction = QtGui.QAction(QtGui.QIcon('./img/exit.png'), 'Exit', self)
#		exitAction.setShortcut('Ctrl+Q')
#		exitAction.triggered.connect(QtGui.qApp.quit)

#		coordinateSystemAction = QtGui.QAction(QtGui.QIcon('./img/coord.png'), 'Move coordinate system', self)
#		coordinateSystemAction.triggered.connect(self.onMoveCoordinateSystem)

#		gaussCreateAction = QtGui.QAction(QtGui.QIcon('./img/create_gauss.png'), 'Create Gaussians', self)
#		gaussCreateAction.triggered.connect(self.onCreateGaussians)

#		gaussModifyAction = QtGui.QAction(QtGui.QIcon('./img/modify_gauss.png'), 'Modify Gaussians', self)
#		gaussModifyAction.triggered.connect(self.onModifyGaussians)

		self.moveCoordinateSystemButton = QtWidgets.QToolButton()
		self.moveCoordinateSystemButton.setIcon(QtGui.QIcon(resource_path('./img/coord.png')))
		self.moveCoordinateSystemButton.setCheckable(True)
		self.moveCoordinateSystemButton.setChecked(True)
		self.moveCoordinateSystemButton.clicked.connect(self.onMoveCoordinateSystem)

		self.createGaussButton = QtWidgets.QToolButton()
		self.createGaussButton.setIcon(QtGui.QIcon(resource_path('./img/create_gauss.png')))
		self.createGaussButton.setCheckable(True)
		self.createGaussButton.clicked.connect(self.onCreateGaussians)

		self.modifyGaussButton = QtWidgets.QToolButton()
		self.modifyGaussButton.setIcon(QtGui.QIcon(resource_path('./img/modify_gauss.png')))
		self.modifyGaussButton.setCheckable(True)
		self.modifyGaussButton.clicked.connect(self.onModifyGaussians)

		self.createSamplesButton = QtWidgets.QToolButton()
		self.createSamplesButton.setIcon(QtGui.QIcon(resource_path('./img/samples.png')))
		self.createSamplesButton.setCheckable(True)
		self.createSamplesButton.clicked.connect(self.onCreateSamples)
		
		self.toolbar =	QtWidgets.QToolBar()
		self.addToolBar(QtCore.Qt.RightToolBarArea, self.toolbar)
		# self.toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
		# self.toolbar.addAction(coordinateSystemAction)
		# self.toolbar.addAction(gaussCreateAction)
		# self.toolbar.addAction(gaussModifyAction)
		self.toolbar.addWidget(self.moveCoordinateSystemButton)
		self.toolbar.addWidget(self.createGaussButton)
		self.toolbar.addWidget(self.modifyGaussButton)
		self.toolbar.addWidget(self.createSamplesButton)

		
	def closeEvent(self, event):
		try:
			self.featurespace.saveDefaultFeatureSpace('.featurespace.pyct')
		except:
			print("could not save default feature space")
			
		event.accept()


	def newFeatureSpace(self):
		self.classifier = None
		self.regressor = None
		self.featurespace.new()


	def openFeatureSpace(self):
		self.classifier = None
		self.regressor = None
		self.featurespace.open()


	def saveFeatureSpace(self):
		self.featurespace.save()


	def saveAsFeatureSpace(self):
		self.featurespace.saveAs()


	def exportFeatureSpace(self):
		self.featurespace.exportFile()


	def exportFeatureSpaceAsImage(self):
		self.featurespace.exportImage()


	def importFeatureSpace(self):
		self.featurespace.importFile()


	def onMoveCoordinateSystem(self):
		self.moveCoordinateSystemButton.setChecked(True)
		self.createGaussButton.setChecked(False)
		self.modifyGaussButton.setChecked(False)
		self.createSamplesButton.setChecked(False)
		self.createSamplesDockWidget.setHidden(True)
		self.featurespace.changeAction(self.featurespace.ACTION_COORDINATE_SYSTEM)
		self.statusbar.showMessage('')


	def onCreateGaussians(self):
		self.moveCoordinateSystemButton.setChecked(False)
		self.createGaussButton.setChecked(True)
		self.modifyGaussButton.setChecked(False)
		self.createSamplesButton.setChecked(False)
		self.createSamplesDockWidget.setHidden(True)
		self.featurespace.changeAction(self.featurespace.ACTION_CREATE_GAUSSIAN)


	def onModifyGaussians(self):
		self.moveCoordinateSystemButton.setChecked(False)
		self.createGaussButton.setChecked(False)
		self.modifyGaussButton.setChecked(True)
		self.createSamplesButton.setChecked(False)
		self.createSamplesDockWidget.setHidden(True)
		self.featurespace.changeAction(self.featurespace.ACTION_MODIFY_GAUSSIAN)
		self.statusbar.showMessage('')

	def onCreateSamples(self):
		self.moveCoordinateSystemButton.setChecked(False)
		self.createGaussButton.setChecked(False)
		self.modifyGaussButton.setChecked(False)
		self.createSamplesButton.setChecked(True)
		self.createSamplesDockWidget.setHidden(False)
		self.featurespace.changeAction(self.featurespace.ACTION_CREATE_SAMPLES)
		self.statusbar.showMessage('')


	def clusterWithParameters(self, method):
		self.clusteringParameters.setTab(method)
		result = self.clusteringParameters.exec_()
		if result == QtWidgets.QDialog.Accepted:
			self.classifier = None
			self.featurespace.setClassificationImage(None)
			self.regressor = None
			self.clusterer = Clustering(method, self.clusteringParameters, self.featurespace)
			try:
				self.clusterer.initialize()
			except AssertionError as e:
				QtWidgets.QMessageBox.warning(self, 'Error', str(e), QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Ok)		
			self.repaint()


	def editClusteringParameters(self):
		self.clusteringParameters.setTab(-1)
		self.clusteringParameters.exec_()


	def reduceDimensionalityWithParameters(self, method):
		self.dimRedParameters.setTab(method)
		result = self.dimRedParameters.exec_()
		if result == QtWidgets.QDialog.Accepted:
			self.classifier = None
			self.featurespace.setClassificationImage(None)
			self.regressor = None
			self.dimreduction = DimensionalityReduction(method, self.dimRedParameters, self.featurespace)
			self.dimreduction.initialize()
			self.repaint()


	def classify(self, classifier):
		self.regressor = None
		self.classifier = Classifier(classifier, self.classifierParameters, self.featurespace)
		try:
			self.classifier.initialize()
			self.runFeatureSpaceComputations()
		except Exception as e:
			QtWidgets.QMessageBox.warning(self, 'Error', str(e), QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Ok)		
			


	def classifyWithParameters(self, classifier):
		self.regressor = None
		self.classifierParameters.setTab(classifier)
		result = self.classifierParameters.exec_()
		if result == QtWidgets.QDialog.Accepted:
			self.classifier = Classifier(classifier, self.classifierParameters, self.featurespace)
			try:
				self.classifier.initialize()
				self.runFeatureSpaceComputations()
			except AssertionError as e:
				QtWidgets.QMessageBox.warning(self, 'Error', str(e), QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Ok)


	def unsetClassifier(self):
		self.classifier = None
		self.featurespace.setClassificationImage(None)
		self.repaint()


	def runFeatureSpaceComputations(self, initialize = False):
		changes = False
		if self.classifier:
			self.setCursor(QtGui.QCursor(QtCore.Qt.WaitCursor))
			self.statusbar.showMessage('')

			img = None
			if initialize:
				try:
					self.classifier.initialize()
					img = self.classifier.runFeatureSpaceComputations()
				except AssertionError as e:
					QtWidgets.QMessageBox.warning(self, 'Error', str(e), QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Ok)
			else:
				img = self.classifier.runFeatureSpaceComputations()

			self.featurespace.setClassificationImage(img)
			self.featurespace.repaint()
			self.statusbar.showMessage('classification done')
			self.setCursor(QtGui.QCursor(QtCore.Qt.ArrowCursor))
			changes = True
		if self.regressor:
			self.statusbar.showMessage('')

			if initialize:
				self.regressor.initialize()

			self.featurespace.repaint()

		return changes


	def editClassificationParameters(self):
		self.classifierParameters.setTab(-1)
		self.classifierParameters.exec_()


	def regression(self, regressor):
		self.setCursor(QtGui.QCursor(QtCore.Qt.WaitCursor))
		self.statusbar.showMessage('')
		self.regressor = Regression(regressor, self.regressionParameters, self.featurespace)
		self.regressor.initialize()
		self.setCursor(QtGui.QCursor(QtCore.Qt.ArrowCursor))
		self.statusbar.showMessage('regression done')
		self.classifier = None
		self.featurespace.setClassificationImage(None)
		self.repaint()


	def regressionWithParameters(self, regressor):
		self.regressionParameters.setTab(regressor)
		result = self.regressionParameters.exec_()
		if result == QtWidgets.QDialog.Accepted:
			self.setCursor(QtGui.QCursor(QtCore.Qt.WaitCursor))
			self.statusbar.showMessage('')
			self.regressor = Regression(regressor, self.regressionParameters, self.featurespace)
			try:
				self.regressor.initialize()
			except AssertionError as e:
				QtWidgets.QMessageBox.warning(self, 'Error', str(e), QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Ok)
			
			self.setCursor(QtGui.QCursor(QtCore.Qt.ArrowCursor))
			self.statusbar.showMessage('regression done')
			self.classifier = None
			self.featurespace.setClassificationImage(None)
			self.repaint()


	def paintRegressor(self, qp):
		if self.regressor:
			self.regressor.paint(qp)


	def unsetRegressor(self):
		self.regressor = None
		self.repaint()


	def editRegressionParameters(self):
		self.regressionParameters.setTab(-1)
		self.regressionParameters.exec_()



def main():
	app = QtWidgets.QApplication(sys.argv)
	win = PyClassificationToolbox()
	win.setWindowTitle("Python Classification Toolbox")
	win.show()	
	sys.exit(app.exec_())


if __name__=="__main__":
	main()

